import sys
import asyncio
import logging
from html import parser as html_parser
from argparse import ArgumentParser

from httpx import AsyncClient, URL

from knoc import __version__
from knoc.settings import settings, KMiweoSettings
from knoc.kmiweo_customer_info import MiweoCustomer, parse_customer_info
from knoc.models.kmiweo import (
    record_types,
    MiweoIdRecord,
    ListParams,
    List,
    ListItem,
)

logger = logging.getLogger("kmiweo")


class KMiweoError(Exception):
    pass


class KMiweoLoginError(KMiweoError):
    pass


class KMiweoApiError(KMiweoError):
    pass


class KMiweoCustomerNotFoundError(KMiweoApiError):
    pass


class MiweoRecordHtmlParser(html_parser.HTMLParser):
    record_id: MiweoIdRecord

    def handle_starttag(self, tag, attrs):
        if tag == "a":
            for attr in attrs:
                if attr[0] == "href" and isinstance(attr[1], str):
                    self.record_id = MiweoIdRecord(**URL(attr[1]).params)
                    return


class HtmlCleanerParser(html_parser.HTMLParser):
    """HTML parser that returns the text (content)"""

    text: str

    def handle_data(self, data):
        self.text = data


def extract_text(html: str) -> str | None:
    """Extract the text from an html string"""
    if html == "-":
        return None
    html_cleaner_parser = HtmlCleanerParser()
    html_cleaner_parser.feed(html)
    return html_cleaner_parser.text.strip()


class MiweoAPI(AsyncClient):
    _miweo_domain: URL
    _miweao_record_html_parser: MiweoRecordHtmlParser
    _conn_params: KMiweoSettings

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._miweao_record_html_parser = MiweoRecordHtmlParser()
        self.set_connection_params(settings.miweo)

    def set_connection_params(self, conn_params: KMiweoSettings):
        self._conn_params = conn_params
        self.cookies.set("miweoCookie", conn_params.cookie)
        self.cookies.set("PHPSESSID", conn_params.phpsessid)

    async def get_customer_info(self, customer_id: str) -> MiweoCustomer:
        """Get customer info for the customer id"""
        customer_record_id = await self.get_customer_record_id(customer_id=customer_id)
        return await self.get_customer_info_from_miweo_id(customer_record_id)

    async def get_customer_record_id(self, customer_id: str) -> MiweoIdRecord:
        """Get internal Miweo id record id for the customer id"""
        resp = await self.get(
            "Ajax/controlerAjaxModule.php",
            params={
                "search": customer_id,
            },
        )
        if resp.is_error:
            raise KMiweoApiError(str(resp))
        resp_json = resp.json()
        try:
            results = resp_json["data"]["results"]
        except KeyError:
            raise KMiweoApiError("Malformed Miweo response (no data.results")
        if len(results) == 0:
            raise KMiweoApiError(f"No customer found with customer_id {customer_id}")
        return self.get_miweo_record_id(results[0])

    async def get_customer_info_from_miweo_id(
        self, customer_record_id: MiweoIdRecord
    ) -> MiweoCustomer:
        # https://knet.miweo.com/index.php?link=35753&id=104157&ajax=t
        resp = await self.get(
            "index.php",
            params={
                "link": customer_record_id.link,
                "id": customer_record_id.id,
                "ajax": "t",
            },
        )
        if resp.is_error:
            raise KMiweoApiError(str(resp))
        return parse_customer_info(resp.text)

    def get_miweo_record_id(self, record) -> MiweoIdRecord:
        """Extract link and id for the record"""
        url = record[-1]
        self._miweao_record_html_parser.feed(url)
        if not hasattr(self._miweao_record_html_parser, "record_id"):
            raise KMiweoApiError("No internal id found in record")
        return self._miweao_record_html_parser.record_id

    def get_list_list(self) -> list[str]:
        """Return the list of lists (record types)"""
        return list(record_types.keys())

    async def get_list(self, request: ListParams) -> List:
        """Get all records, querying the "table" page several times"""
        items: list[ListItem] = []
        page_number = 1
        while True:
            partial = await self.get_list_page(request, page_number)
            items.extend(partial.items)
            if len(items) >= partial.length:
                break
            page_number += 1
        return List(items=items, length=partial.length)

    async def get_list_page(self, request: ListParams, page: int) -> List:
        params = request.model_dump()
        params["page"] = page
        del params["url"]
        resp = await self.get(request.url, params=params)
        if resp.is_error:
            raise KMiweoApiError(str(resp))
        resp_json = resp.json()
        data = resp_json["data"]["results"]
        # Return only the first field of each line: the id
        return List(
            items=[
                ListItem(
                    id=line[request.id_column],
                    status=(
                        extract_text(line[request.status_column])
                        if request.status_column
                        else None
                    ),
                )
                for line in data
            ],
            length=resp_json["data"]["totalResultsNumber"],
        )

    # async def get_first(self, request: ListParams) -> List:


miweo_api = MiweoAPI(base_url=URL(str(settings.miweo.url)))


async def main():
    parser = ArgumentParser("KMiweo CLI")
    parser.add_argument("-c", "--customer-id")
    parser.add_argument("-k", "--cookie")
    parser.add_argument("-p", "--phpsessid")
    parser.add_argument("-l", "--list")
    parser.add_argument("-L", "--lists", action="store_true")
    parser.add_argument(
        "-v", "--version", action="store_true", help="Print knoc version and exits"
    )
    args = parser.parse_args()
    if args.version:
        print(__version__)
        sys.exit()
    logging.basicConfig(level=logging.INFO)
    if not (args.cookie and args.phpsessid):
        logger.error(
            "The cookie and phpsessid must be provided "
            "(get those from a valid session on a browser)"
        )
        sys.exit(1)
    kmiweo_settings = KMiweoSettings(cookie=args.cookie, phpsessid=args.phpsessid)
    miweo_api.set_connection_params(kmiweo_settings)
    if args.lists:
        print("\n".join(miweo_api.get_list_list()))
    elif args.list:
        if args.list not in record_types:
            logger.warning(f"Unknown list {args.list}")
            sys.exit(2)
        result = await miweo_api.get_list(record_types[args.list])
        logger.info(f"Got {len(result.items)} / {result.length} items")
        print("\n".join([f"{item.id};{item.status}" for item in result.items]))
    elif args.customer_id:
        print(await miweo_api.get_customer_info(customer_id=args.customer_id))
    else:
        print("Nothing to do")


if __name__ == "__main__":
    asyncio.run(main())
