from datetime import timedelta

from pydantic import BaseModel

import pandas as pd  # type: ignore
from hatchet_sdk import Context, EmptyModel

from knoc import __version__
from knoc.settings import settings
from knoc.db import aio_df_to_sql, aio_sql_to_df
from knoc.hatchet_client import hatchet
from knoc.kmiweo import miweo_api, record_types
from knoc.workflows.utils import ImportReport, ImportReportRecordType


class MiweoImportParams(BaseModel):
    to_db: bool = True
    to_feather: bool = False
    record_types: list[str] = []


import_records_wf = hatchet.workflow(
    name="Import-Miweo-Customers",
    input_validator=MiweoImportParams,
    version=__version__,
)

consolidate_records_wf = hatchet.workflow(
    name="Consolidate-Miweo-records", version=__version__
)


@import_records_wf.task(
    execution_timeout=timedelta(minutes=5), schedule_timeout=timedelta(minutes=10)
)
async def import_miweo_records(
    input: MiweoImportParams,
    ctx: Context,
) -> ImportReport:
    """Import miweo customers lists.
    If lists is None, get all lists."""
    report = ImportReport()
    if len(input.record_types) == 0:
        input.record_types = list(record_types.keys())
    for record_type in input.record_types:
        ctx.log(f"Getting {record_type}...")
        result = await miweo_api.get_list(record_types[record_type])
        df = pd.DataFrame(result.model_dump()["items"])
        if input.to_db:
            ctx.log(f"Write to DB {len(df)} records")
            await aio_df_to_sql(
                df,
                settings.miweo.db_table_records_list_prefix + record_type,
                schema=settings.miweo.db_schema,
                if_exists="replace",
            )
        if input.to_feather:
            df.to_feather(settings.data_path / "customers.feather")
        report.types.append(ImportReportRecordType(name=record_type, count=len(df)))
    ## Consolidate data if all record types have been imported
    if input.record_types == list(record_types.keys()):
        await consolidate_records_wf.aio_run(EmptyModel())
        report.message = "OK, and all data consolidated"
    return report


@consolidate_records_wf.task()
async def consolidate_miweo_records(
    input: EmptyModel,
    ctx: Context,
):
    """Import miweo customers lists.
    If lists is None, get all lists."""
    assert input
    dfs: dict[str, pd.DataFrame] = {}

    for record_type in list(record_types.keys()):
        df = await aio_sql_to_df(
            settings.miweo.db_table_records_list_prefix + record_type,
            schema=settings.miweo.db_schema,
        )
        dfs[record_type] = df
        ## Write the table with all records, adding a column "type"
        for recort_type, df in dfs.items():
            df["record_type"] = recort_type

    df_all = pd.concat(dfs.values())
    ctx.log(f"Write to DB {len(df_all)} records")
    await aio_df_to_sql(
        df_all,
        settings.miweo.db_table_records_list,
        schema=settings.miweo.db_schema,
        if_exists="replace",
    )
